/* SD Karten Interface */

#include "StdTypes.h"
#include "STM32F103.h"
#include "startup.h"
#include "sdio.h"
#include "systick.h"
#include "events.h"
#include "serial.h"
#include "conv.h"

#define  ComPort  2

/* verwendete Portbits:
 PG.8   WPI    low = Kontakt geschlossen
 PG.7   CDI    low = SD-Karte vorhanden
 */

inline bool is_WPI (void) { return ((GPIOG->IDR & (1<<8))!=0); }
inline bool is_CDI (void) { return ((GPIOG->IDR & (1<<7))==0); }

/* Variablen */
dword CSD_Tab[4];    /* Card Specific Data       */
dword CID_Tab[4];    /* Card Identification Data */
dword RCA;           /* momentane Card Adresse   */
dword SCR;           /* SDCard Conf Register     */
dword SCR2;          /* OEM SDCard Conf Register */
dword OCR;           /* mit ACMD41 gelesen       */
dword TRSTATUS;      /* letzter Status bei Rd+Wr */

TSDCARDINFO SDCardInfo;

bool SD_debounce;

// extern  dword MCI_ReadFifo(dword * dest);
// extern  dword MCI_WriteFifo(dword * src);


/*************** zyklischer Test, ob SD Karte eingesetzt wurde ***************/
int SD_exdebounce;

void SDCardTest (void)
{ dword L;

  if (SD_CardDetected)              // wenn Karte bereits erkannt wurde
  { if (SD_debounce) return;        // erstmal entprellen lassen
    if (!is_CDI())                  // kein SDI im Betrieb mehr? --> Karte wurde entnommen!
    { ++SD_exdebounce;
      if (SD_exdebounce > 3)
      { SDCardTyp = 0;              // Typ austragen
        SD_CardDetected = false;
        SD_readonly = false;
        SDIO_POWER = 0;
        Add_Event(idSDverloren);
      }
    }
    else SD_exdebounce = 0;
    return;
  }
  if (!is_CDI()) return;            // erstmal entprellen lassen
  if (!SD_CardDetected) Add_Delayed_Event(200, idSDentdeckt);  // nur 1x Event erzeugen
  SD_CardDetected = true;
  L = 10000;
  while (--L);
  SD_readonly = is_WPI();
  SD_debounce = true;
  SD_exdebounce = 0;
}

/*************** Debug- und Testfunktionen ***********************************/
void Sage4 (dword* ptr)
{ HexL_Out(ptr[0],ComPort);  Char_Out('-',ComPort);
  HexL_Out(ptr[1],ComPort);  Char_Out('-',ComPort);
  HexL_Out(ptr[2],ComPort);  Char_Out('-',ComPort);
  HexL_Out(ptr[3],ComPort);
  CRLF_Out(ComPort);
}



/*************** Kommandos & Antworten ***************************************/
// responses:
#define Rbusy    (1<<16)
#define Rapp     (1<<17)
#define NoCRC    (1<<18)

#define Resp0    (0)            // keine Antwort
#define Resp1    (1<<6)         // 48 bit
#define Resp1b   (1<<6)|Rbusy   // 48 bit
#define Resp2    (3<<6)         // 136 bit CID oder CSD
#define Resp3    (1<<6)|NoCRC   // 48 bit  OCD
#define Resp4    (1<<6)         // erstmal...  ist NUR fr SDIO Karten
#define Resp6    (1<<6)         // 48 bit  RCA
#define Resp7    (1<<6)         // 48 bit  supported Voltages


#define  CMD0     0|Resp0            // reset, gehe in Idle Mode
#define  CMD1     1|Resp1            // nur MMC: sende OpCond Register
#define  CMD2     2|Resp2            // sende CID Nummer
#define  CMD3     3|Resp6            // sende RCA und gehe in Data Transfer Mode
#define  CMD4     4|Resp0            // programmiert DSR
#define  CMD5     5|Resp4            // nur fr SDIO-Karten
#define ACMD6     6|Resp1|Rapp       // setze Busbreite (1 oder 4)
#define  CMD7     7|Resp1b           // select/deselect Karte
#define  CMD8     8|Resp7            // sende Interface Conds
#define  CMD9     9|Resp2            // sende CSD
#define  CMD10   10|Resp2            // sende CID
#define  CMD11   11|Resp1            // umschalten auf 1.8 V Signalpegel
#define  CMD12   12|Resp1            // stop Transmission
#define  CMD13   13|Resp1            // sende Status
#define ACMD13   13|Resp1|Rapp       // sende Status
#define  CMD15   15|Resp0            // deaktiviere Karte
#define  CMD16   16|Resp1            // setze Blocklnge (512 normalerweise)
#define  CMD17   17|Resp1            // read single block
#define  CMD18   18|Resp1            // read multiple blocks
#define  CMD19   19|Resp1            // send tuning block pattern
#define  CMD20   20|Resp1b           // speed class control
#define ACMD22   22|Resp1|Rapp       // sende Anzahl fehlerfrei geschriebener Blcke
#define  CMD23   23|Resp1            // setze Block-Count fr Cmd 18 + 25
#define ACMD23   23|Resp1|Rapp       // setze Anzahl der zu lschenden Blcke
#define  CMD24   24|Resp1            // write block
#define  CMD25   25|Resp1            // write multiple blocks
#define  CMD27   27|Resp1            // programmiert Teile von CSD
#define  CMD28   28|Resp1b           // setze Write-Protect (nicht bei SDHC, SDXC)
#define  CMD29   29|Resp1b           // lsche Write-Protect (nicht bei SDHC, SDXC)
#define  CMD30   30|Resp1            // sende Write-Protect Status (nicht bei SDHC, SDXC)
#define  CMD32   32|Resp1            // setze Adresse fr das Lschen des 1. Write Blockes
#define  CMD33   33|Resp1            // setze Adresse fr das Lschen des letzten Write Blockes
#define  CMD38   38|Resp1b
#define  CMD40   40|Resp1
#define ACMD41   41|Resp3|Rapp       // OpConds (HCS/OCR), NICHT bei MMC, NICHT bei SDIO
#define  CMD42   42|Resp1
#define ACMD42   42|Resp1|Rapp
#define  CMD48   48|Resp1
#define  CMD49   49|Resp1
#define ACMD51   51|Resp1|Rapp
#define  CMD55   55|Resp1            // leitet ein ACMD-Kommando ein
#define  CMD56   56|Resp1
#define  CMD58   58|Resp1
#define  CMD59   59|Resp1

// statusbits, statisch
#define CCRCFAIL           1
#define DCRCFAIL       (1<<1)
#define CTIMEOUT       (1<<2)
#define DTIMEOUT       (1<<3)
#define TXUNDERR       (1<<4)
#define RXOVERR        (1<<5)
#define CMDRESPEND     (1<<6)
#define CMDSENT        (1<<7)
#define DATAEND        (1<<8)
#define STBITERR       (1<<9)
#define DBCKEND        (1<<10)

// statusbits, dynamisch
#define CMDACTIV       (1<<11)
#define TXACTIV        (1<<12)
#define RXACTIV        (1<<13)
#define TXFIFOHE       (1<<14)
#define RXFIFOHF       (1<<15)
#define TXFIFOF        (1<<16)
#define RXFIFOF        (1<<17)
#define TXFIFOE        (1<<18)
#define RXFIFOE        (1<<19)


bool SD_WaitForResponse (dword Command)
{ dword L;

  L = Ticks + 2;                               // maximal 2 ms warten Cortex-> Systick verwenden!
  if (Command & (1<<6))                       // wenn Response erwartet wird warten wir
                                              // auf CMDRESPEND oder CMDCRCFAIL
  { while (((SDIO_STA & (CMDRESPEND | CCRCFAIL | CTIMEOUT))==0) && (Ticks <= L));
    if (SDIO_STA & CMDRESPEND) return true; // bei CmdRespEnd --> OK.
    if (Command & NoCRC)
    { if (SDIO_STA & CCRCFAIL) return true; }
    return false;
  }
                                              // kein Response erwartet, also warten auf CMDSENT
  while ((SDIO_STA & (CMDSENT | CTIMEOUT))==0)
      if (Ticks > L) return false;
  if (SDIO_STA & CMDSENT) return true;
  return false;
}


/* Kommando zur SD Karte ausgeben
   ==============================
Format:  0000 0ePI LRKK KKKK
0..5  = Kommando
6     = ob Response
7     = Response ist lang
8     = Interrupt?
9     = Pending?
10    = enable
11..31 --> 0
*/

bool SD_Command (dword Command, dword Argument)
{ bool b;

  if (Command & Rapp)                        // bei allen ACMDxx --> CMD55 davor
  { SDIO_CMD  = 0;                           // erstmal stopp
    SDIO_ICR  = 0x7FF;
    SDIO_ARG  = RCA<<16;
    SDIO_CMD  = CMD55 | (1<<10);             //  starten und enable
    b = SD_WaitForResponse(Command);
    if (b) b = (SDIO_RESP1 & (1<<5)) != 0;    // ob auch APP_CMD im Statuswort gesetzt ist
    if (!b) return false;
  }
  SDIO_CMD  = 0;                             // erstmal stopp
  SDIO_ICR    = 0x7FF;
  SDIO_ARG = Argument;
  SDIO_CMD  = (Command & 0x7FF) | (1<<10);   // starten und enable
  if (!SD_WaitForResponse (Command)) return false;
  return true;
}

int Get_SD_Status (void)
{ bool b;
  int i;
  b = SD_Command(CMD13, RCA<<16); // sende Status
  if (!b) return -1;
  i = (SDIO_RESP1 >> 9) & 15;
  return i;
}



/*************** Anschlu und SD-Karte initialisieren ************************/
bool SD_DriveInit (void)
{ bool  b;
  bool  v2compliant;
  dword i, L, irca;
  int64 isize;

  SDIO_POWER  = SDIO_CLKCR  =
  SDIO_ARG    = SDIO_CMD    =
  SDIO_DTIMER = SDIO_DLEN   = 0;
  SDIO_ICR    = 0x00C007FF;

  SD_CardSize = 0;
  Warte_ms(100);
  SD_readonly = is_WPI();
  SD_debounce = false;

  RCC->AHBENR |= (1<<10);         // SD-System einschalten
  SDIO_POWER   = 2 | (1<<6);      // Powerup + CMD=Opendrain
  Warte_ms(200);

  SDIO_POWER = 3 | (1<<6);        // Poweron + CMD=Opendrain

/* HCLK = 72MHz, SDIOCLK = 72MHz, SDIO_CK = HCLK/(178 + 2) = 400 KHz */
  SDIO_CLKCR = 178 |(1<<14) |(1<<8);
  SDIO_CMD   = 0;                 // Command Maschine disable
  SDIO_ICR   = 0x00C007FF;        // alle bisherigen Fehler lschen
  SDIO_DCTRL = 0;                 // Datentransfer disable
  SDIO_MASK  = 0;                 // erstmal keine Interrupts
  Warte_ms(500);

  SDCardTyp = SD_err;
  RCA = 0;

/*
  VICVectAddr24 = (dword)MCI_IRQHandler;  // hat  Slot 24
  VICIntSelect  &=  ~(1<<24);             // kein FIQ
  VICVectPriority28 = 1;                  // Prioritt 1 (zweithchste)
  VICIntEnable  = (1<<24);                // enable MCI im VIC
*/

  /********* Stufe 0: CMD0 = Resetkommando **********/
  b = SD_Command(CMD0, 0);       // CMD0 = Reset
  if (!b) return false;

  /********* Stufe 1: CMD1 = Test auf MMC  **********/
  String_Out("\r\nCMD1 ",ComPort);
  b = SD_Command(CMD1, 0);
  if ((b) && (SDIO_RESP1 & 0x80000000))
  { SDCardTyp = MMC;
    String_Out("--> MMC\r\n",ComPort);

    // erstmal nix bei MMC's

    return false;
  }
  String_Out("no\r\n",ComPort);  // kein CMD1 --> nix oder SD Karte

  /********* Stufe 2: CMD8 = Test auf SD Version 2.0 oder neuer **********/
  String_Out("Cmd8 ",ComPort);
  b = SD_Command(CMD8, 0x1AA);   // CMD8 = send Interface Conds (Test auf R2.0 = SDHC)
  v2compliant = b && (SDIO_RESP1==0x1AA);
  if (v2compliant)
    String_Out("SD2.0\r\n",ComPort);
  else
    String_Out("SD1.x\r\n",ComPort);


  /********* Stufe 3a: SD 1.1 initialisieren **********/
  if (!v2compliant)
  { b = SD_Command(ACMD41, 0);   // ohne HCS, ohne 1.8V
    if (!b) return false;        // MMC oder Karte defekt oder fr falsche VCC

    String_Out("SD1.x OCR= ",ComPort);
    OCR = 0;
    L = Ticks + 1000;            // 1 Sekunde max.
    while ((L>Ticks) && ((OCR & 0x80000000)==0))
    { b = SD_Command(ACMD41, 0x00100000);  // VCC 3.2V bis 3.3V
      if (!b) return false;
      OCR = SDIO_RESP1;
      Warte_ms(3);
    }
    if (OCR & 0x80000000)
    { SDCardTyp = SD1_1;         // SD Standard Card 1.1
      goto _init1_1;
    }
    String_Out(" ..timeout\r\n",ComPort);
    return false;
  }


  /********* Stufe 3b: SD 2.0 initialisieren **********/
  OCR = 0;                     // SD ist 2.0 oder hher, wir machen erst mal nur 2.0
  L = Ticks + 1000;            // 1 Sekunde max.
  while ((L>Ticks) && ((OCR & 0x80000000)==0))
  { b = SD_Command(ACMD41, 0x40100000);  // HCS und VCC 3.2V bis 3.3V
    if (!b) return false;
    OCR = SDIO_RESP1;
    Warte_ms(3);
  }
  if (OCR & 0x80000000)
  { SDCardTyp = SD2_0;         // SD Standard Card 2.0
    goto _init2_0;
  }
  String_Out(" ..timeout\r\n",ComPort);

  SD_Error:
  SDCardTyp = SD_err;
  return false;


  /********* Stufe 3c: Test auf SDHC **********/
  _init2_0:
  if (OCR & 0x40000000)
  { SDCardTyp = SDHC;
    String_Out("SDHC\r\n",ComPort);
  }


  /********* Stufe 4: SD 1.x oder 2.0 initialisieren **********/
  _init1_1:                    // ab hier auch init SD 1.1 Karte

  b = SD_Command(CMD2, 0);     // CID lesen von 'allen' Karten
  CID_Tab[0] = SDIO_RESP1;
  CID_Tab[1] = SDIO_RESP2;
  CID_Tab[2] = SDIO_RESP3;
  CID_Tab[3] = SDIO_RESP4;

  String_Out("CID= ",ComPort);
  Sage4(CID_Tab);

  b = SD_Command(CMD3, 0);     // RCA lesen und in Datenmodus gehen
  if (!b) goto SD_Error;

  irca = (SDIO_RESP1 >> 16) & 0xFFFF;
  RCA = irca;
  String_Out("RCA= ",ComPort);
  HexW_Out(RCA,ComPort);
  CRLF_Out(ComPort);

  b = SD_Command(CMD9, irca<<16);  // CSD lesen
  if (!b) goto SD_Error;

  CSD_Tab[0] = SDIO_RESP1;     // 127..96
  CSD_Tab[1] = SDIO_RESP2;     // 95..64
  CSD_Tab[2] = SDIO_RESP3;     // 63..32
  CSD_Tab[3] = SDIO_RESP4;     // 31..0

  String_Out("CSD= ",ComPort);
  Sage4(CSD_Tab);

  b = SD_Command(CMD7, RCA<<16);   // Karte selektieren
  if (!b) goto SD_Error;

  b = SD_Command(CMD16, 8);        // Blocklnge auf 8 Bytes setzen
  if (!b) goto SD_Error;

  SCR  = 0;
  SCR2 = 0;
  SDIO_DTIMER = 0x10000;
  SDIO_DLEN   = 8;
  SDIO_DCTRL  = 1 | 2 | (3<<4);              // enable, SD-->Controller, 8 Byte
  b = SD_Command(ACMD51, 0);                 // SCR lesen
  if (!b) goto SD_Error;
  while ((SDIO_STA & ((1<<3)|(1<<8))) == 0); // wartet auf DataEnd oder DataTimeout
  if (SDIO_STA & (1<<21))                    // wenn Daten im RX Fifo
  { SCR = SDIO_FIFO;
    String_Out("SCR  = ",ComPort);
    HexL_Out(SCR,ComPort);
  }
  if (SDIO_STA & (1<<21))                    // wenn Daten im RX Fifo
  { SCR2 = SDIO_FIFO;
    String_Out("\r\nSCR2 = ",ComPort);
    HexL_Out(SCR2,ComPort);
  }


  /********* Stufe 5: bergang zu 4 Bit  **********/
  SDIO_POWER   = 3;            // Poweron ohne OpenDrain
  i = Get_SD_Status();
  String_Out("\r\nStatus = ",ComPort);
  Dezi_Out(i,1,ComPort);
  CRLF_Out(ComPort);
  if (i==5) SD_Command(CMD12, 0);

  String_Out("ACMD6  ",ComPort);
  b = SD_Command(ACMD6, 2);   // auf 4 Bit Transfer bergehen
  if (!b) goto SD_Error;

  String_Out("ok\r\n",ComPort);

  // Clock und Datenbreite umschalten (72MHz SDCARD_CLK)
  // MCICLK = SDCARD_CLK / (divisor + 2);

  SDIO_CLKCR = 6 | (1<<11) | (1<<8);

  // Test...
  L = SDIO_CLKCR & 255;
  L = 72000000 / (L+2);
  L = L / 1000;
  Dezi_Out(L,1,ComPort);
  String_Out("kHz/4Bit ",ComPort);
  b = SD_Command(CMD13, RCA<<16); // sende Status
  if (!b) goto SD_Error;

  String_Out("ok\r\nStatus: ",ComPort);
  HexL_Out(SDIO_RESP1,ComPort);
  String_Out("\r\nTyp: ",ComPort);
  switch (SDCardTyp)
  { case SDHC:  String_Out("SDHC\r\n",ComPort); break;
    case SD2_0: String_Out("SD2.0\r\n",ComPort); break;
    case SD1_1: String_Out("SD1.1\r\n",ComPort); break;
    default:    String_Out("sonstige\r\n",ComPort); break;
  }

  // Kapazitt berechnen
  if (CSD_Tab[0] & 0xC0000000)
  { L = (CSD_Tab[1] & 0x3F) << 16;
    L = L | ((CSD_Tab[2] >> 16) & 0xFFFF);
    L = L + 1;
    isize = L;
    isize = isize * 512 * 1024;
  }
  else
  { L = (CSD_Tab[1] & 0x3FF) << 2;
    L = L | ((CSD_Tab[2] >> 30) & 3);
    L = L + 1;
    isize = L;

    L = (CSD_Tab[1] >> 16) & 15;
    L = 1<<L;

    i = (CSD_Tab[2] >> 15) & 7;
    i = 4<<i;
    isize = isize * i * L;
  }
  String_Out("\r\nSize: ",ComPort);
  Dezi_Out(isize/1048576,1,ComPort);
  String_Out(" MB (",ComPort);
  Dezi_Out(isize,1,ComPort);
  String_Out(" Bytes)\r\n",ComPort);
  SD_CardSize = isize;

  String_Out("So, die Karte luft..\r\n>",ComPort);
  return true;
}


/*************** Interrupt ******************************/
// wird erstmal garnicht gebraucht.



/*************** Lese- und Schreiboperationen ******************************/

bool SD_ReadBlock (dword SecNum, dword *readbuff)
{ long  i;
  bool  b;
  dword A;
  dword* BP;

  BP = readbuff;

  if (SDCardTyp <= 0) return false;

  SDIO_ICR = 0x00C007FF;
  SDIO_DCTRL = 0;                    // Data Maschine erstmal aus

  i = Get_SD_Status();
  if (i==5) SD_Command(CMD12, 0);
  if (i<0) return false;

  b = SD_Command(CMD16, 512);        // Blocklnge auf 512 Bytes setzen
  if (!b)
  { TRSTATUS = SDIO_STA;
    SD_Command(CMD12, 0);
    return false;
  }

  i = 512;
  A = SecNum;
  if (SDCardTyp != SDHC) A = A * 512;

  SDIO_MASK     = 0;
  SDIO_DTIMER   = 0x10000;
  SDIO_DLEN     = 512;
  SDIO_DCTRL    = 1 | 2 | (9<<4);     // enable, SD-->Controller, 512 Byte
  b = SD_Command(CMD17, A);           // read single Block
  if (!b) return false;

  readloop:
    A = SDIO_STA;
    if (A & (1<<15))                  // Fifo halbvoll
    { *BP++ = SDIO_FIFO;
      *BP++ = SDIO_FIFO;
      *BP++ = SDIO_FIFO;
      *BP++ = SDIO_FIFO;
      *BP++ = SDIO_FIFO;
      *BP++ = SDIO_FIFO;
      *BP++ = SDIO_FIFO;
      *BP++ = SDIO_FIFO;
      i = i - 32;
      A = SDIO_STA;
    }

    if (A & ((1<<1)|(1<<3)|(1<<5)|(1<<9)))     // crc, timeout, overrun, startbit
    { TRSTATUS = A;
      SD_Command(CMD12, 0);
      return false;
    }

    if (A & ((1<<8)|(1<<10)))                  // dataend, blockdataend
    { while ((SDIO_STA & (1<<21))&&(i>0))
      { *BP++ = SDIO_FIFO;
        i = i - 4;
      }
      return true;
    }

  goto readloop;
}


bool SD_WriteBlock   (dword SecNum, dword *writebuff)
{
  if (SDCardTyp <= 0) return false;

  return true;
}
